#
# This file is part of Python package: `pyxarr`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 SRON
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Unit tests for the pyxarr function `dset_from_h5`."""

from __future__ import annotations

import h5py
import numpy as np


from pyxarr.dset_from_h5 import dset_from_h5


def test_1d() -> None:
    """Check read of a 1-D dataset."""
    flname = "xarr_test_1d.h5"
    # ++++++++++++++++++++++++++++++++++++++++++++++++++
    with h5py.File(flname, "w") as fid:
        fid.create_dataset("array_1d", data=np.random.rand(128).astype(float))

    with h5py.File(flname, "r") as fid:
        print("# Test 1-D, no dims", dset_from_h5(fid["array_1d"]))

    # ++++++++++++++++++++++++++++++++++++++++++++++++++
    with h5py.File(flname, "w") as fid:
        dset = fid.create_dataset("time", data=np.arange(128).astype("f8"))
        dset.attrs["long_name"] = "timestamp of housekeeping data"
        dset.attrs["standard_name"] = "time"
        dset.attrs["calendar"] = "proleptic_gregorian"
        dset.attrs["coverage_content_type"] = "coordinate"
        dset.attrs["units"] = "seconds since 2025-07-28 00:00:00"
        dset.attrs.create("valid_min", 0, dtype="f8")
        dset.attrs.create("valid_max", 92400, dtype="f8")
        dset = fid.create_dataset("array_1d", data=np.random.rand(128).astype(float))
        dset.dims[0].attach_scale(fid["time"])
        
    with h5py.File(flname, "r") as fid:
        print("\n# Test 1-D, with dims", dset_from_h5(fid["array_1d"]))

    # ++++++++++++++++++++++++++++++++++++++++++++++++++
    with h5py.File(flname, "w") as fid:
        dset = fid.create_dataset("time", data=np.arange(128).astype("u4"))
        dset.attrs["long_name"] = "timestamp of housekeeping data"
        dset.attrs["standard_name"] = "time"
        dset.attrs["calendar"] = "proleptic_gregorian"
        dset.attrs["coverage_content_type"] = "coordinate"
        dset.attrs["units"] = "days since 2025-07-01"
        dset = fid.create_dataset("array_1d", data=np.random.rand(128).astype(float))
        dset.dims[0].attach_scale(fid["time"])
        
    with h5py.File(flname, "r") as fid:
        print("\n# Test 1-D, with dims", dset_from_h5(fid["array_1d"]))

    flname = "xarr_test_3d.h5"
    # ++++++++++++++++++++++++++++++++++++++++++++++++++
    with h5py.File(flname, "w") as fid:
        dset = fid.create_dataset("time", data=np.arange(128).astype("u4"))
        dset.attrs["long_name"] = "timestamp of housekeeping data"
        dset.attrs["standard_name"] = "time"
        dset.attrs["calendar"] = "proleptic_gregorian"
        dset.attrs["coverage_content_type"] = "coordinate"
        dset.attrs["units"] = "days since 2025-07-01"
        dset = fid.create_dataset("row", data=np.arange(11).astype("u2"))
        dset = fid.create_dataset("column", data=np.arange(17).astype("u2"))
        dset = fid.create_dataset(
            "array_3d",
            data=np.random.rand((128 * 11 * 17)).astype(float).reshape(128, 11, 17),
        )
        dset.dims[0].attach_scale(fid["time"])
        dset.dims[1].attach_scale(fid["row"])
        dset.dims[2].attach_scale(fid["column"])
        
    with h5py.File(flname, "r") as fid:
        print("\n# Test 3-D, with dims", dset_from_h5(fid["array_3d"]))



if __name__ == "__main__":
    test_1d()



#
# This file is part of Python package: `pyxarr`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 SRON
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of pyxarr function `dset_from_h5`."""

from __future__ import annotations

__all__ = ["dset_from_h5"]

from pathlib import PurePath
from typing import TYPE_CHECKING

import numpy as np

from .xarr import Coord, DataArray

if TYPE_CHECKING:
    import h5py


# - local functions --------------------------------
def __get_attrs(dset: h5py.Dataset, field: str) -> dict:
    """Return attributes of the HDF5 dataset.

    Parameters
    ----------
    dset :  h5py.Dataset
       HDF5 dataset from which the attributes are read
    field : str
       Name of field in compound dataset

    Returns
    -------
    dict with numpy arrays

    """
    _field = None
    if field is not None:
        try:
            _field = {
                "name": field,
                "oneof": len(dset.dtype.names),
                "index": dset.dtype.names.index(field),
            }
        except Exception as exc:
            raise RuntimeError(f"field {field} not in dataset {dset.name}") from exc
        # print('_field ', _field)

    attrs = {}
    for key in dset.attrs:
        if key in (
            "CLASS",
            "DIMENSION_LIST",
            "NAME",
            "REFERENCE_LIST",
            "_Netcdf4Dimid",
            "_Netcdf4Coordinates",
        ):
            continue

        attr_value = dset.attrs[key]
        # print('# ----- ', key, type(attr_value), attr_value)
        if isinstance(attr_value, np.ndarray):
            if len(attr_value) == 1:
                attr_value = attr_value[0]
                # print('# ----- ', key, type(attr_value), attr_value)
            elif _field is not None and len(attr_value) == _field["oneof"]:
                attr_value = attr_value[_field["index"]]
                # elif isinstance(attr_value, np.void):
                #    attr_value = attr_value[0]

        attrs[key] = (
            attr_value.decode("ascii") if isinstance(attr_value, bytes) else attr_value
        )

    return attrs


def __get_coords(dset: h5py.Dataset, data_sel: tuple[slice | int]) -> list:
    r"""Return coordinates of the HDF5 dataset with dimension scales.

    Parameters
    ----------
    dset :  h5py.Dataset
       HDF5 dataset from which the data is read
    data_sel :  tuple of slice or int
       A numpy slice generated for example `numpy.s\_`

    Returns
    -------
    A sequence of tuples [(name, data), ...]

    """
    coords = []
    if len(dset.dims) == dset.ndim:
        try:
            for ii, dim in enumerate(dset.dims):
                # get name of dimension
                name = PurePath(dim[0].name).name
                if name.startswith("row") or name.startswith("column"):
                    name = name.split(" ")[0]

                # determine coordinate
                buff = None
                if dim[0].size > 0:
                    if np.all(dim[0][()] == 0):
                        d_type = "u2" if ((dset.shape[ii] - 1) >> 16) == 0 else "u4"
                        buff = np.arange(dset.shape[ii], dtype=d_type)
                    else:
                        buff = dim[0][()]

                if not (buff is None or data_sel is None):
                    buff = buff[data_sel[ii]]

                coords.append((name, buff))
        except RuntimeError as exc:
            raise RuntimeError(
                f"failed to collect coordinates of dataset {dset.name}"
            ) from exc

    return coords


def __set_coords(
    dset: h5py.Dataset | np.ndarray,
    data_sel: tuple[slice | int] | None,
    dim_names: list | None,
) -> list:
    r"""Set coordinates of the HDF5 dataset.

    Parameters
    ----------
    dset :  h5py.Dataset or np.ndarray
       HDF5 dataset from which the data is read, or numpy array
    data_sel :  tuple of slice or int
       A numpy slice generated for example `numpy.s\_`
    dim_names : list of strings
       Alternative names for the dataset dimensions if not attached to dataset
       Default coordinate names are ['time', ['row', ['column']]]

    Returns
    -------
    A sequence of tuples [(name, data), ...]

    """
    if dim_names is None:
        if dset.ndim > 3:
            raise ValueError("not implemented for ndim > 3")

        dim_names = ["time", "row", "column"][-dset.ndim :]

    coords = []
    for ii in range(dset.ndim):
        co_dtype = "u2" if ((dset.shape[ii] - 1) >> 16) == 0 else "u4"
        buff = np.arange(dset.shape[ii], dtype=co_dtype)
        if data_sel is not None:
            buff = buff[data_sel[ii]]
        coords.append((dim_names[ii], buff))

    return coords


def __get_data(
    dset: h5py.Dataset, data_sel: tuple[slice | int] | None, field: str
) -> np.ndarray:
    r"""Return data of the HDF5 dataset.

    Parameters
    ----------
    dset :  h5py.Dataset
       HDF5 dataset from which the data is read
    data_sel :  tuple of slice or int
       A numpy slice generated for example `numpy.s\_`
    field : str
       Name of field in compound dataset or None

    Returns
    -------
    Numpy array

    Notes
    -----
    Read floats always as doubles

    """
    if data_sel is None:
        data_sel = ()

    if np.issubdtype(dset.dtype, np.floating):
        data = dset.astype(float)[data_sel]
        data[data == float.fromhex("0x1.ep+122")] = np.nan
        return data

    if field is None:
        return dset[data_sel]

    data = dset.fields(field)[data_sel]
    if np.issubdtype(data.dtype, np.floating):
        data = data.astype(float)
        data[data == float.fromhex("0x1.ep+122")] = np.nan
    return data


def __check_selection(data_sel: slice | tuple | int, ndim: int) -> slice | tuple | None:
    r"""Check and correct user provided data selection.

    Notes
    -----
    If data_sel is used to select data from a dataset then the number of
    dimensions of data_sel should agree with the HDF5 dataset or one and
    only one Ellipsis has to be used.
    Thus allowed values for data_sel are:
    * [always]: (), np.s\_[:], np.s\_[...]
    * [1-D dataset]: np.s\_[:-1], np.s\_[0]
    * [2-D dataset]: np.s\_[:-1, :], np.s\_[0, :], np.s\_[:-1, 0]
    * [3-D dataset]: np.s\_[:-1, :, 2:4], np.s\_[0, :, :], np.s\_[:-1, 0, 2:4]
    * [Ellipsis] np.s\_[0, ...], np.s\_[..., 4], np.s\_[0, ..., 4]

    """
    if data_sel in (np.s_[:], np.s_[...], np.s_[()]):
        return None

    if np.isscalar(data_sel):
        return np.s_[data_sel : data_sel + 1]

    buff = ()
    for val in data_sel:
        if val == Ellipsis:
            for _ in range(ndim - len(data_sel) + 1):
                buff += np.index_exp[:]
        elif np.isscalar(val):
            buff += (np.s_[val : val + 1],)
        else:
            buff += (val,)

    return buff


# - main function ----------------------------------
def dset_from_h5(
    h5_dset: h5py.Dataset,
    data_sel: tuple[slice | int] | None = None,
    *,
    dim_names: list[str] | None = None,
    field: str | None = None,
) -> DataArray:
    r"""Create pyxarr.DataArray from a HDF5 dataset (with dimension scales).

    Implements a lite interface with the pyxarr.DataArray, should work for all
    2-D detector images, sequences of detector measurements and trend data.

    Parameters
    ----------
    h5_dset :  h5py.Dataset
       Data, dimensions, coordinates and attributes are read for this dataset
    data_sel :  tuple of slice or int, optional
       A numpy slice generated for example `numpy.s\_`
    dim_names :  list of strings, optional
       Alternative names for the dataset dimensions if not attached to dataset
    field : str, optional
       Name of field in compound dataset or None

    Returns
    -------
    pyxarr.DataArray

    Notes
    -----
    All floating datasets are converted to Python type 'float'

    Dimensions and Coordinates:

    * The functions in this module should work with netCDF4 and HDF5 files.
    * In a HDF5 file the 'coordinates' of a dataset can be defined using \
      dimension scales.
    * In a netCDF4 file this is required: all variables have dimensions, \
      which can have coordinates. But under the hood also netCDF4 uses \
      dimension scales.
    * The DataArray structure will have as dimensions, the names of \
      the dimension scales and as coordinates the names and data of the \
      dimensions scales, except when the data only contains zero's.
    * The default dimensions of an image are 'row' and 'column' with evenly \
      spaced values created with np.arange(len(dim), dtype=uint).

    Examples
    --------
    Read HDF5 dataset 'signal' from file::

    > fid = h5py.File(flname, 'r')        # flname is a HDF5/netCDF4 file
    > xdata = dset_from_h5(fid['signal'])
    > fid.close()

    """
    # Check data selection
    if data_sel is not None:
        data_sel = __check_selection(data_sel, h5_dset.ndim)
    # print(f"data_sel: {data_sel}")

    # get coordinates of the dataset
    coords = []
    if dim_names is None:
        coords = __get_coords(h5_dset, data_sel)
    if not coords:
        coords = __set_coords(h5_dset, data_sel, dim_names)
    # print(f"dim_names: {dim_names}")
    # print(f"coords: {coords}, {len(coords)}")

    # get values for the dataset
    data = __get_data(h5_dset, data_sel, field)
    # print(f"data: {data.ndim}, {data.shape}")

    # - check if dimension of dataset and coordinates agree
    if data.ndim < len(coords):
        for ii in reversed(range(len(coords))):
            if np.isscalar(coords[ii][1]):
                del coords[ii]

    # - remove empty coordinates
    co_tuple = tuple(Coord(key, val) for key, val in coords if val is not None)
    # print(f"co_tuple: {co_tuple}")

    # get dataset attributes
    attrs = __get_attrs(h5_dset, field)

    # get name of the dataset
    name = PurePath(h5_dset.name).name if field is None else field

    return DataArray(data, coords=co_tuple, name=name, attrs=attrs)



#
# This file is part of Python package: `pyxarr`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 SRON
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of the pyxarr class `DataArray`."""

from __future__ import annotations

__all__ = ["Coord", "DataArray"]

from dataclasses import KW_ONLY, dataclass, field
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

# - global parameters ------------------------------
MAX_DIMS = 3


# - class Coord ------------------------------------
@dataclass(frozen=True)
class Coord:
    """Define one coordinate of a labeled array.

    Parameters
    ----------
    name :  str, optional
      name of the coordinate
    values :  ArrayLike, optional
      values or labels of the coordinate

    """

    name: str | None = None
    values: ArrayLike | None = None

    def __bool__(self: Coord) -> bool:
        """Return False if Coord is empty."""
        return self.name is not None


# - class DataArray --------------------------------
@dataclass()
class DataArray:
    """Define multi-dimensional labeled array.

    Parameters
    ----------
    values :  ArrayLike, optional
      multi-dimensional array
    coords : tuple[Coord, ...], optional
      coordinates for each dimension of the array
    dims :  tuple[str, ...], optional
      dimension names for each dimension of the array
    attrs :  dict[str, Any], optional
      attributes providing meta-data of the array
    name :  str, optional
      name of the array

    """

    values: ArrayLike | None = None
    _: KW_ONLY
    coords: tuple[Coord, ...] = field(default_factory=tuple)
    dims: tuple[str, ...] = field(default_factory=tuple)
    attrs: dict = field(default_factory=dict)
    name: str | None = None

    def __post_init__(self: DataArray) -> None:
        """Initialize undefined class variables of DataArray."""
        if self.values is None:
            return

        # make sure that values are a numpy array
        self.values = np.asarray(self.values)

        # define coordinates
        if not self.coords:
            if self.values.ndim > MAX_DIMS:
                raise KeyError("Please provide coordinates of N-dim > 3")
            if self.dims and len(self.dims) != self.values.ndim:
                raise KeyError("Please provide dimension names for each dimension")

            # use dimension names or use default dimension names
            _keys = (
                self.dims
                if self.dims
                else ("time", "row", "column")[MAX_DIMS - self.values.ndim :]
            )
            for ii in range(self.values.ndim):
                self.coords += (Coord(_keys[ii], list(range(self.values.shape[ii]))),)
        else:
            if len(self.coords) != self.values.ndim:
                raise KeyError("Please provide coordinates for each dimension")

        # define dimensions
        if not self.dims:
            for coord in self.coords:
                self.dims += (coord.name,)

        # add attributes units and long_name
        # for key in ["units", "long_name"]:
        #    if key not in self.attrs:
        #        self.attrs[key] = ""

    def __bool__(self: DataArray) -> bool:
        """Return False if DataArray is empty."""
        return self.values is not None


def tests() -> None:
    """..."""
    attrs = {
        "long_name": "my dataset",
        "units": "m/s",
    }

    print("# --- empty data array ---")
    xarr = DataArray(attrs=attrs)
    if xarr:
        print(xarr.values)
        print(xarr.coords)
        print(xarr.dims)
    print(xarr.attrs)

    print("# --- 1-D array without coordinates ---")
    xarr = DataArray(list(range(11)), attrs=attrs)
    if xarr:
        print(xarr.values)
        print(xarr.coords)
        print(xarr.dims)
        print(xarr.attrs)

    print("\n# --- 2-D array without coordinates ---")
    xarr = DataArray(np.arange(3 * 7).reshape(3, 7))
    print(xarr.values)
    print(xarr.coords)
    print(xarr.dims)
    print(xarr.attrs)

    print("\n# --- 3-D array without coordinates ---")
    xarr = DataArray(np.arange(5 * 3 * 7).reshape(5, 3, 7))
    print(xarr.values)
    print(xarr.coords)
    print(xarr.dims)
    print(xarr.attrs)

    print("\n# --- 2-D array without coordinates with dimension names ---")
    xarr = DataArray(np.arange(3 * 7).reshape(3, 7), dims=("x", "y"))
    print(xarr.values)
    print(xarr.coords)
    print(xarr.dims)
    print(xarr.attrs)

    print("\n# --- 2-D array with coordinates ---")
    xarr = DataArray(
        np.arange(3 * 7).reshape(3, 7),
        coords=(Coord("x", [1, 2, 3]), Coord("y", np.arange(7))),
    )
    print(xarr.values)
    print(xarr.coords)
    print(xarr.dims)
    print(xarr.attrs)


if __name__ == "__main__":
    tests()



#
# This file is part of Python package: `pyxarr`
#
#     https://github.com/rmvanhees/pyxarr.git
#
# Copyright (c) 2025 SRON
#    All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of pyxarr function `dset_from_xr`."""

from __future__ import annotations

__all__ = ["dset_from_xr"]

from typing import TYPE_CHECKING

from .xarr import Coord, DataArray

if TYPE_CHECKING:
    import xarray as xr


# - local functions --------------------------------


# - main function ----------------------------------
def dset_from_xr(xda: xr.DataArray) -> DataArray:
    """Copy content of xarray.DataArray to a pyxarr DataArray.

    Parameters
    ----------
    xda : xr.DataArray
       the xarray.DataArray to be copied to a pyxarr.DataArray

    """
    ds_coords = tuple(Coord(x, xda.coords[x]) for x in xda.coords)

    return DataArray(xda.values, coords=ds_coords, attrs=xda.attrs, name=xda.name)
